#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for CubicComplementarityVLE formulation

Authors: Andrew Lee
"""

import pytest

from pyomo.environ import (
    ConcreteModel,
    Constraint,
    Expression,
    Param,
    value,
    Var,
    units as pyunits,
)

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.phase_equil.smooth_VLE_2 import (
    CubicComplementarityVLE,
    _calculate_temperature_slacks,
    _calculate_ceos_derivative_slacks,
    EPS_INIT,
)
from idaes.models.properties.modular_properties.eos.ideal import Ideal
from idaes.models.properties.modular_properties.eos.ceos import Cubic, CubicType
from idaes.models.properties.modular_properties.phase_equil.forms import fugacity
from idaes.core.util.exceptions import ConfigurationError


@pytest.mark.unit
def test_different_cubics():
    m = ConcreteModel()

    m.params = GenericParameterBlock(
        components={
            "H2O": {
                "parameter_data": {
                    "pressure_crit": (220.6e5, pyunits.Pa),
                    "temperature_crit": (647, pyunits.K),
                    "omega": 0.344,
                },
                "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            }
        },
        phases={
            "Liq": {
                "equation_of_state": Cubic,
                "equation_of_state_options": {"type": CubicType.PR},
            },
            "Vap": {
                "equation_of_state": Cubic,
                "equation_of_state_options": {"type": CubicType.SRK},
            },
        },
        state_definition=FTPx,
        pressure_ref=100000.0,
        temperature_ref=300,
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        phases_in_equilibrium=[("Vap", "Liq")],
        phase_equilibrium_state={("Vap", "Liq"): CubicComplementarityVLE},
        parameter_data={
            "PR_kappa": {
                ("H2O", "H2O"): 0.000,
            },
            "SRK_kappa": {
                ("H2O", "H2O"): 0.000,
            },
        },
    )

    with pytest.raises(
        ConfigurationError,
        match="params - CubicComplementarityVLE formulation requires that both phases use the same "
        "type of cubic equation of state.",
    ):
        m.props = m.params.state_block_class([1], parameters=m.params)


@pytest.mark.unit
def test_non_cubic():
    m = ConcreteModel()

    m.params = GenericParameterBlock(
        components={
            "H2O": {
                "parameter_data": {
                    "pressure_crit": (220.6e5, pyunits.Pa),
                    "temperature_crit": (647, pyunits.K),
                    "omega": 0.344,
                },
                "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            }
        },
        phases={
            "Liq": {
                "equation_of_state": Cubic,
                "equation_of_state_options": {"type": CubicType.PR},
            },
            "Vap": {
                "equation_of_state": Ideal,
            },
        },
        state_definition=FTPx,
        pressure_ref=100000.0,
        temperature_ref=300,
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        phases_in_equilibrium=[("Vap", "Liq")],
        phase_equilibrium_state={("Vap", "Liq"): CubicComplementarityVLE},
        parameter_data={
            "PR_kappa": {
                ("H2O", "H2O"): 0.000,
            }
        },
    )

    with pytest.raises(
        ConfigurationError,
        match="params - CubicComplementarityVLE formulation only supports cubic equations of state.",
    ):
        m.props = m.params.state_block_class([1], parameters=m.params)


@pytest.fixture()
def frame():
    m = ConcreteModel()

    # Create a dummy parameter block
    m.params = GenericParameterBlock(
        components={
            "H2O": {
                "parameter_data": {
                    "pressure_crit": (220.6e5, pyunits.Pa),
                    "temperature_crit": (647, pyunits.K),
                    "omega": 0.344,
                },
                "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            }
        },
        phases={
            "Liq": {
                "equation_of_state": Cubic,
                "equation_of_state_options": {"type": CubicType.PR},
            },
            "Vap": {
                "equation_of_state": Cubic,
                "equation_of_state_options": {"type": CubicType.PR},
            },
        },
        state_definition=FTPx,
        pressure_ref=100000.0,
        temperature_ref=300,
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        phases_in_equilibrium=[("Vap", "Liq")],
        phase_equilibrium_state={("Vap", "Liq"): CubicComplementarityVLE},
        parameter_data={
            "PR_kappa": {
                ("H2O", "H2O"): 0.000,
            }
        },
    )

    # Create a dummy state block
    m.props = m.params.state_block_class([1], parameters=m.params)

    return m


@pytest.mark.unit
def test_build(frame):
    assert isinstance(frame.props[1].s_Vap_Liq, Var)
    assert isinstance(frame.props[1].gp_Vap_Liq, Var)
    assert isinstance(frame.props[1].gn_Vap_Liq, Var)

    assert isinstance(frame.props[1].cubic_second_derivative_Vap_Liq, Expression)

    assert isinstance(frame.props[1].eps_t_Vap_Liq, Param)
    assert value(frame.props[1].eps_t_Vap_Liq) == pytest.approx(1, rel=1e-8)
    assert isinstance(frame.props[1].eps_z_Vap_Liq, Param)
    assert value(frame.props[1].eps_z_Vap_Liq) == pytest.approx(1, rel=1e-8)

    assert isinstance(frame.props[1]._teq_constraint_Vap_Liq, Constraint)
    assert isinstance(
        frame.props[1].temperature_slack_complementarity_Vap_Liq, Constraint
    )
    assert isinstance(frame.props[1].cubic_slack_complementarity_Vap_Liq, Constraint)

    for k in ["Liq", "Vap"]:
        assert k in frame.props[1].s_Vap_Liq
        assert k in frame.props[1].gp_Vap_Liq
        assert k in frame.props[1].gn_Vap_Liq
        assert k in frame.props[1].cubic_second_derivative_Vap_Liq
        assert k in frame.props[1].temperature_slack_complementarity_Vap_Liq
        assert k in frame.props[1].cubic_slack_complementarity_Vap_Liq


# TODO: Need tests for formulation


@pytest.mark.unit
def test_cubic_second_derivative(frame):
    Z = frame.props[1].compress_fact_phase

    # For pure water
    R = 8.314
    b = 0.07780 * R * 647 / 220.6e5

    for P in range(1, 11):
        for T in range(300, 500, 10):
            frame.props[1].pressure.set_value(P * 1e5)
            frame.props[1].temperature.set_value(T)
            frame.props[1]._teq[("Vap", "Liq")].set_value(T)

            B = b * P * 1e5 / R / T

            for p in ["Vap", "Liq"]:
                der = value(6 * Z[p] + 2 * -(1 + B - 2 * B))
                assert value(
                    frame.props[1].cubic_second_derivative_Vap_Liq[p]
                ) == pytest.approx(der, rel=1e-8)


@pytest.mark.unit
def test_calculate_temperature_slacks(frame):
    s = frame.props[1].s_Vap_Liq
    # Teq > T
    frame.props[1].temperature.set_value(300)
    frame.props[1]._teq[("Vap", "Liq")].set_value(400)
    _calculate_temperature_slacks(frame.props[1], ("Vap", "Liq"), "Liq", "Vap")

    assert value(s["Vap"]) == pytest.approx(100, rel=1e-8)
    assert value(s["Liq"]) == pytest.approx(EPS_INIT, abs=1e-8)

    # Teq < T
    frame.props[1]._teq[("Vap", "Liq")].set_value(200)
    _calculate_temperature_slacks(frame.props[1], ("Vap", "Liq"), "Liq", "Vap")

    assert value(s["Liq"]) == pytest.approx(100, rel=1e-8)
    assert value(s["Vap"]) == pytest.approx(EPS_INIT, abs=1e-8)

    # Teq == T
    frame.props[1]._teq[("Vap", "Liq")].set_value(300)
    _calculate_temperature_slacks(frame.props[1], ("Vap", "Liq"), "Liq", "Vap")

    assert value(s["Liq"]) == pytest.approx(EPS_INIT, abs=1e-8)
    assert value(s["Vap"]) == pytest.approx(EPS_INIT, abs=1e-8)


@pytest.mark.unit
def test_calculate_ceos_derivative_slacks(frame):
    gp = frame.props[1].gp_Vap_Liq
    gn = frame.props[1].gn_Vap_Liq
    der = frame.props[1].cubic_second_derivative_Vap_Liq
    Z = frame.props[1].compress_fact_phase

    # For pure water
    R = 8.314
    b = 0.07780 * R * 647 / 220.6e5

    # Atmospheric conditions, vapor derivative +ve, liquid -ve
    frame.props[1].pressure.set_value(1e5)
    frame.props[1].temperature.set_value(300)
    frame.props[1]._teq[("Vap", "Liq")].set_value(300)

    _calculate_ceos_derivative_slacks(frame.props[1], ("Vap", "Liq"), "Liq", "Vap")

    B = value(b * frame.props[1].pressure / R / frame.props[1].temperature)
    der_l = value(6 * Z["Liq"] + 2 * -(1 + B - 2 * B))
    der_v = value(6 * Z["Vap"] + 2 * -(1 + B - 2 * B))

    assert value(gp["Liq"]) == pytest.approx(EPS_INIT, abs=1e-8)
    assert value(gn["Liq"]) == pytest.approx(-der_l, rel=1e-8)

    assert value(gp["Vap"]) == pytest.approx(der_v, rel=1e-8)
    assert value(gn["Vap"]) == pytest.approx(EPS_INIT, abs=1e-8)

    # Supercritical conditions, vapor derivative +ve, liquid +ve
    frame.props[1].pressure.set_value(250e5)
    frame.props[1].temperature.set_value(700)
    frame.props[1]._teq[("Vap", "Liq")].set_value(700)

    _calculate_ceos_derivative_slacks(frame.props[1], ("Vap", "Liq"), "Liq", "Vap")

    B = value(b * frame.props[1].pressure / R / frame.props[1].temperature)
    der_v = value(6 * Z["Vap"] + 2 * -(1 + B - 2 * B))

    assert value(gp["Liq"]) == pytest.approx(EPS_INIT, abs=1e-8)
    assert value(gn["Liq"]) == pytest.approx(EPS_INIT, abs=1e-8)

    assert value(gp["Vap"]) == pytest.approx(der_v, rel=1e-8)
    assert value(gn["Vap"]) == pytest.approx(EPS_INIT, abs=1e-8)

    # Supercritical conditions, vapor derivative -ve, liquid -ve
    frame.props[1].pressure.set_value(250e5)
    frame.props[1].temperature.set_value(300)
    frame.props[1]._teq[("Vap", "Liq")].set_value(300)

    _calculate_ceos_derivative_slacks(frame.props[1], ("Vap", "Liq"), "Liq", "Vap")

    B = value(b * frame.props[1].pressure / R / frame.props[1].temperature)
    der_l = value(6 * Z["Liq"] + 2 * -(1 + B - 2 * B))

    assert value(gp["Liq"]) == pytest.approx(EPS_INIT, abs=1e-8)
    assert value(gn["Liq"]) == pytest.approx(-der_l, rel=1e-8)

    assert value(gp["Vap"]) == pytest.approx(EPS_INIT, abs=1e-8)
    assert value(gn["Vap"]) == pytest.approx(EPS_INIT, abs=1e-8)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Unit operation model for a steam heater applicable to platen superheater
and roof superheater, model main equations:

* Heat is given by fire-side boiler model
* Calculate pressure change due to friction
* Calculate slag layer wall temperature
* Consider a layer of metal and a layer of slag

Created on Aug 27, 2020 by Boiler Team (J. Ma, M. Zamarripa)
"""
import pytest

# Import Pyomo libraries
import pyomo.environ as pyo

# Import IDAES core
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom

# Import Unit Model Modules
from idaes.models.properties import iapws95

from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver
from idaes.models_extra.power_generation.properties import FlueGasParameterBlock
from idaes.models_extra.power_generation.unit_models.boiler_fireside import (
    BoilerFireside,
)
from idaes.models_extra.power_generation.unit_models.tests.datadictionary import (
    data_dic,
)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------


@pytest.fixture(scope="module")
def build_unit():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()
    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(dynamic=False)
    # Add property packages to flowsheet library
    m.fs.prop_fluegas = FlueGasParameterBlock()
    # boiler based on surrogate
    m.fs.unit = BoilerFireside(
        dynamic=False,
        property_package=m.fs.prop_fluegas,
        calculate_PA_SA_flows=False,
        number_of_zones=12,
        has_platen_superheater=True,
        has_roof_superheater=True,
        surrogate_dictionary=data_dic,
    )
    return m


@pytest.mark.unit
def test_basic_build(build_unit):
    """Make a boiler fire side model
    and make sure it doesn't throw exception"""
    m = build_unit
    assert degrees_of_freedom(m) == 29
    # Check unit config arguments
    assert len(m.fs.unit.config) == 9
    assert m.fs.unit.config.number_of_zones == 12
    assert isinstance(m.fs.unit.waterwall_heat, pyo.Var)
    assert isinstance(m.fs.unit.platen_heat, pyo.Var)
    assert isinstance(m.fs.unit.roof_heat, pyo.Var)
    assert m.fs.unit.config.property_package is m.fs.prop_fluegas


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize_unit(build_unit):
    m = build_unit
    # platen superheater
    # fix coal ultimate analysis on dry basis
    # using Coal Illinois No. 6 bituminous coal
    m.fs.unit.mf_C_coal_dry.fix(0.6300)
    m.fs.unit.mf_H_coal_dry.fix(0.0412)
    m.fs.unit.mf_O_coal_dry.fix(0.1300)
    m.fs.unit.mf_N_coal_dry.fix(0.0127)
    m.fs.unit.mf_S_coal_dry.fix(0.0284)
    m.fs.unit.mf_Ash_coal_dry.fix(0.1577)
    m.fs.unit.hhv_coal_dry.fix(2.581e007)
    m.fs.unit.frac_moisture_vaporized[:].fix(0.6)
    m.fs.unit.mf_H2O_coal_raw[:].fix(0.2)  # moisture content
    m.fs.unit.flowrate_coal_raw[:].fix(44.0)  # kg/s

    m.fs.unit.wall_temperature_waterwall[:, :].fix(690)
    m.fs.unit.wall_temperature_platen[:].fix(750)
    m.fs.unit.wall_temperature_roof[:].fix(650)
    m.fs.unit.fcorrection_heat_ww.fix(0.95)
    m.fs.unit.fcorrection_heat_platen.fix(0.95)
    # SCPC simulation approx flue gas = 21290.6999  # mol/s
    flow_mol_pa = 21290.6999 * 0.34  # approx. 1/3 as Primary air
    flow_mol_sa = 21290.6999 * 0.66  # approx. 2/3 as Secondary air

    state_args_PA = {
        "flow_mol_comp": {
            "H2O": 0.0078267 * flow_mol_pa,
            "CO2": 0.000337339 * flow_mol_pa,
            "N2": 0.783994 * flow_mol_pa,
            "O2": 0.20784 * flow_mol_pa,
            "SO2": 1e-5 * flow_mol_pa,
            "NO": 1e-5 * flow_mol_pa,
        },
        "temperature": 333.15,
        "pressure": 101325.00,
    }
    state_args_SA = {
        "flow_mol_comp": {
            "H2O": 0.0078267 * flow_mol_sa,
            "CO2": 0.000337339 * flow_mol_sa,
            "N2": 0.783994 * flow_mol_sa,
            "O2": 0.20784 * flow_mol_sa,
            "SO2": 1e-5 * flow_mol_sa,
            "NO": 1e-5 * flow_mol_sa,
        },
        "temperature": 650.15,
        "pressure": 101325.00,
    }

    for i in m.fs.prop_fluegas.component_list:
        m.fs.unit.primary_air_inlet.flow_mol_comp[:, i].fix(
            state_args_PA["flow_mol_comp"][i]
        )
        m.fs.unit.secondary_air_inlet.flow_mol_comp[:, i].fix(
            state_args_SA["flow_mol_comp"][i]
        )
    m.fs.unit.primary_air_inlet.pressure[:].fix(101325.00)
    m.fs.unit.secondary_air_inlet.pressure[:].fix(101325.00)
    m.fs.unit.primary_air_inlet.temperature[:].fix(333.15)
    m.fs.unit.secondary_air_inlet.temperature[:].fix(650.15)
    m.fs.unit.temperature_coal[:].fix(335.15)
    m.fs.unit.SR_lf.fix(1.0)
    m.fs.unit.deltaP.fix(1000)
    assert degrees_of_freedom(m) == 0

    initialization_tester(
        build_unit, dof=0, state_args_PA=state_args_PA, state_args_SA=state_args_SA
    )


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solve_unit(build_unit):
    m = build_unit
    assert degrees_of_freedom(m) == 0

    results = solver.solve(m, tee=True)
    # Check for optimal solution
    assert pyo.check_optimal_termination(results)

    # flue gas outlet temperature
    assert pytest.approx(1236.780228, abs=1e-3) == pyo.value(
        m.fs.unit.flue_gas_outlet.temperature[0]
    )

    # check material balance
    assert (
        pytest.approx(
            pyo.value(
                m.fs.unit.primary_air_moist[0].flow_mass
                + m.fs.unit.secondary_air[0].flow_mass
                + m.fs.unit.flowrate_coal_burner[0]
                - m.fs.unit.flue_gas[0].flow_mass
                - m.fs.unit.flow_mass_flyash[0]
            ),
            abs=1e-2,
        )
        == 0
    )

    #     # check energy balance
    assert (
        pytest.approx(
            pyo.value(
                (
                    m.fs.unit.primary_air_moist[0].flow_mol
                    * m.fs.unit.primary_air_moist[0].enth_mol
                    + m.fs.unit.secondary_air[0].flow_mol
                    * m.fs.unit.secondary_air[0].enth_mol
                    + m.fs.unit.h_coal[0] * m.fs.unit.flowrate_coal_burner[0]
                )
                - (
                    sum(m.fs.unit.waterwall_heat[0, j] for j in m.fs.unit.zones)
                    + (
                        m.fs.unit.platen_heat[0]
                        if m.fs.unit.config.has_platen_superheater is True
                        else 0
                    )
                    + (
                        m.fs.unit.roof_heat[0]
                        if m.fs.unit.config.has_roof_superheater is True
                        else 0
                    )
                    + m.fs.unit.flue_gas[0].flow_mol * m.fs.unit.flue_gas[0].enth_mol
                    + m.fs.unit.flowrate_ash[0]
                    * (
                        593 * (m.fs.unit.flue_gas[0].temperature - 298.15)
                        + 0.293
                        * (m.fs.unit.flue_gas[0].temperature ** 2.0 - 298.15**2.0)
                    )
                    + m.fs.unit.flowrate_daf_flyash[0] * m.fs.unit.hs_daf_flyash[0]
                )
            ),
            abs=1e-3,
        )
        == 0
    )


@pytest.fixture(scope="module")
def build_unit_option2():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()
    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(dynamic=False)
    # Add property packages to flowsheet library
    m.fs.prop_fluegas = FlueGasParameterBlock()
    # boiler based on surrogate
    m.fs.unit = BoilerFireside(
        dynamic=False,
        property_package=m.fs.prop_fluegas,
        calculate_PA_SA_flows=True,
        number_of_zones=12,
        has_platen_superheater=True,
        has_roof_superheater=True,
        surrogate_dictionary=data_dic,
    )
    return m


@pytest.mark.unit
def test_build_option2(build_unit_option2):
    """Make a boiler fire side model with option 2
    and make sure it doesn't throw exception"""
    m = build_unit_option2
    assert degrees_of_freedom(m) == 19
    # Check unit config arguments
    assert len(m.fs.unit.config) == 9
    assert m.fs.unit.config.number_of_zones == 12
    assert isinstance(m.fs.unit.waterwall_heat, pyo.Var)
    assert isinstance(m.fs.unit.platen_heat, pyo.Var)
    assert isinstance(m.fs.unit.SR, pyo.Var)
    assert isinstance(m.fs.unit.ratio_PA2coal, pyo.Var)
    assert isinstance(m.fs.unit.roof_heat, pyo.Var)
    assert m.fs.unit.config.property_package is m.fs.prop_fluegas



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Implementation of the formulation proposed in:

Burgard, A.P., Eason, J.P., Eslick, J.C., Ghouse, J.H., Lee, A., Biegler, L.T.,
Miller, D.C., 2018, A Smooth, Square Flash Formulation for Equation-Oriented
Flowsheet Optimization. Proceedings of the 13th International Symposium on
Process Systems Engineering – PSE 2018, July 1-5, 2018, San Diego.
"""
# TODO: Pylint complains about variables with _x names as they are built by other classes
# pylint: disable=protected-access

from pyomo.environ import Constraint, Param, Var, value
from idaes.core.util.math import smooth_max, smooth_min
from idaes.models.properties.modular_properties.base.utility import (
    identify_VL_component_list,
)
import idaes.core.util.scaling as iscale


# -----------------------------------------------------------------------------
class SmoothVLE(object):
    """Methods for constructing equations associated with Smooth VLE formulation."""

    @staticmethod
    def phase_equil(b, phase_pair):
        """
        Method for constructing phase equilibrium variables and constraints
        """
        # This method is called via StateBlock.build, thus does not need clean-up
        # try/except statements
        suffix = "_" + phase_pair[0] + "_" + phase_pair[1]

        # Smooth VLE assumes a liquid and a vapor phase, so validate this
        (
            _,
            _,
            _,
            _,
            l_only_comps,
            v_only_comps,
        ) = identify_VL_component_list(b, phase_pair)

        # Definition of equilibrium temperature for smooth VLE
        t_units = b.params.get_metadata().default_units.TEMPERATURE
        if v_only_comps == []:
            b.add_component(
                "_t1" + suffix,
                Var(
                    initialize=b.temperature.value,
                    doc="Intermediate temperature for calculating Teq",
                    units=t_units,
                ),
            )
            _t1 = getattr(b, "_t1" + suffix)

            b.add_component(
                "eps_1" + suffix,
                Param(
                    default=0.01,
                    mutable=True,
                    doc="Smoothing parameter for Teq",
                    units=t_units,
                ),
            )
            eps_1 = getattr(b, "eps_1" + suffix)

            # PSE paper Eqn 13
            def rule_t1(b):
                return _t1 == smooth_max(
                    b.temperature, b.temperature_bubble[phase_pair], eps_1
                )

            b.add_component("_t1_constraint" + suffix, Constraint(rule=rule_t1))
        else:
            _t1 = b.temperature

        if l_only_comps == []:
            b.add_component(
                "eps_2" + suffix,
                Param(
                    default=0.0005,
                    mutable=True,
                    doc="Smoothing parameter for Teq",
                    units=t_units,
                ),
            )
            eps_2 = getattr(b, "eps_2" + suffix)

            # PSE paper Eqn 14
            # TODO : Add option for supercritical extension
            def rule_teq(b):
                return b._teq[phase_pair] == smooth_min(
                    _t1, b.temperature_dew[phase_pair], eps_2
                )

        elif v_only_comps == []:

            def rule_teq(b):
                return b._teq[phase_pair] == _t1

        else:

            def rule_teq(b):
                return b._teq[phase_pair] == b.temperature

        b.add_component("_teq_constraint" + suffix, Constraint(rule=rule_teq))

    @staticmethod
    def calculate_scaling_factors(b, phase_pair):
        """
        Method to calculate scaling factors for phase equilibrium
        """
        suffix = "_" + phase_pair[0] + "_" + phase_pair[1]
        sf_T = iscale.get_scaling_factor(b.temperature, default=1, warning=True)

        try:
            _t1 = getattr(b, "_t1" + suffix)
            _t1_cons = getattr(b, "_t1_constraint" + suffix)
            iscale.set_scaling_factor(_t1, sf_T)
            iscale.constraint_scaling_transform(_t1_cons, sf_T, overwrite=False)
        except AttributeError:
            pass

        _teq_cons = getattr(b, "_teq_constraint" + suffix)
        iscale.constraint_scaling_transform(_teq_cons, sf_T, overwrite=False)

    @staticmethod
    def phase_equil_initialization(b, phase_pair):
        """
        Method to initialize phase equilibrium
        """
        suffix = "_" + phase_pair[0] + "_" + phase_pair[1]

        for c in b.component_objects(Constraint):
            # Activate equilibrium constraints
            if c.local_name in ("_t1_constraint" + suffix, "_teq_constraint" + suffix):
                c.activate()

    @staticmethod
    def calculate_teq(b, phase_pair):
        """
        Method to calculate initial guess for equilibrium temperature
        """
        suffix = "_" + phase_pair[0] + "_" + phase_pair[1]

        if hasattr(b, "eq_temperature_bubble"):
            _t1 = getattr(b, "_t1" + suffix)
            _t1.set_value(
                max(value(b.temperature), value(b.temperature_bubble[phase_pair]))
            )
        else:
            _t1 = b.temperature

        if hasattr(b, "eq_temperature_dew"):
            b._teq[phase_pair].set_value(
                min(value(_t1), value(b.temperature_dew[phase_pair]))
            )
        else:
            b._teq[phase_pair].set_value(value(_t1))



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""This module provides a list of supported component and Pyomo expressions for
some properties not implemented as external functions.
"""

__author__ = "John Eslick"

from idaes.models.properties.general_helmholtz.components.registry import (
    register_helmholtz_component,
    remove_component,
    registered_components,
    viscosity_available,
    thermal_conductivity_available,
    surface_tension_available,
    component_registered,
    clear_component_registry,
    eos_reference,
    viscosity_reference,
    thermal_conductivity_reference,
    surface_tension_reference,
)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for ControlVolumeBlockData.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, Constraint, Expression, Set, units, Var
from pyomo.util.check_units import assert_units_consistent, assert_units_equivalent
from pyomo.common.config import ConfigBlock
from idaes.core import (
    ControlVolume0DBlock,
    ControlVolumeBlockData,
    FlowsheetBlockData,
    declare_process_block_class,
    FlowDirection,
    MaterialBalanceType,
    EnergyBalanceType,
)
from idaes.core.util.exceptions import (
    BalanceTypeNotSupportedError,
    ConfigurationError,
    PropertyNotSupportedError,
)
from idaes.core.util.testing import (
    PhysicalParameterTestBlock,
    ReactionParameterTestBlock,
)
import idaes.logger as idaeslog


# -----------------------------------------------------------------------------
# Mockup classes for testing
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("CVFrame")
class CVFrameData(ControlVolume0DBlock):
    def build(self):
        super(ControlVolumeBlockData, self).build()


# -----------------------------------------------------------------------------
# Basic tests
@pytest.mark.unit
def test_base_build():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    assert len(m.fs.cv.config) == 7
    assert m.fs.cv.config.dynamic is False
    assert m.fs.cv.config.has_holdup is False
    assert m.fs.cv.config.property_package == m.fs.pp
    assert isinstance(m.fs.cv.config.property_package_args, ConfigBlock)
    assert len(m.fs.cv.config.property_package_args) == 0
    assert m.fs.cv.config.reaction_package is None
    assert isinstance(m.fs.cv.config.reaction_package_args, ConfigBlock)
    assert len(m.fs.cv.config.reaction_package_args) == 0
    assert m.fs.cv.config.auto_construct is False

    assert hasattr(m.fs.config, "time")


# -----------------------------------------------------------------------------
# Test add_geometry
@pytest.mark.unit
def test_add_geometry():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.cv.add_geometry()

    assert hasattr(m.fs.cv, "volume")
    assert len(m.fs.cv.volume) == 1.0
    assert m.fs.cv.volume[0].value == 1.0


# -----------------------------------------------------------------------------
# Test add_state_blocks
@pytest.mark.unit
def test_add_state_blocks():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)

    assert hasattr(m.fs.cv, "properties_in")
    assert len(m.fs.cv.properties_in[0].config) == 3
    assert m.fs.cv.properties_in[0].config.defined_state is True
    assert m.fs.cv.properties_in[0].config.has_phase_equilibrium is False
    assert m.fs.cv.properties_in[0].config.parameters == m.fs.pp

    assert hasattr(m.fs.cv, "properties_out")
    assert len(m.fs.cv.properties_out[0].config) == 3
    assert m.fs.cv.properties_out[0].config.defined_state is False
    assert m.fs.cv.properties_out[0].config.has_phase_equilibrium is False
    assert m.fs.cv.properties_out[0].config.parameters == m.fs.pp


@pytest.mark.unit
def test_add_state_block_forward_flow():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.cv.add_state_blocks(
        information_flow=FlowDirection.forward, has_phase_equilibrium=False
    )

    assert m.fs.cv.properties_in[0].config.defined_state is True
    assert m.fs.cv.properties_out[0].config.defined_state is False


@pytest.mark.unit
def test_add_state_block_backward_flow():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.cv.add_state_blocks(
        information_flow=FlowDirection.backward, has_phase_equilibrium=False
    )

    assert m.fs.cv.properties_in[0].config.defined_state is False
    assert m.fs.cv.properties_out[0].config.defined_state is True


@pytest.mark.unit
def test_add_state_blocks_has_phase_equilibrium():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)

    assert m.fs.cv.properties_in[0].config.has_phase_equilibrium is True
    assert m.fs.cv.properties_out[0].config.has_phase_equilibrium is True


@pytest.mark.unit
def test_add_state_blocks_no_has_phase_equilibrium():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_state_blocks()


@pytest.mark.unit
def test_add_state_blocks_custom_args():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()

    m.fs.cv = ControlVolume0DBlock(
        property_package=m.fs.pp, property_package_args={"test": "test"}
    )

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)

    assert len(m.fs.cv.properties_in[0].config) == 4
    assert m.fs.cv.properties_in[0].config.test == "test"

    assert len(m.fs.cv.properties_out[0].config) == 4
    assert m.fs.cv.properties_out[0].config.test == "test"


# -----------------------------------------------------------------------------
# Test add_reaction_blocks
@pytest.mark.unit
def test_add_reaction_blocks():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    assert hasattr(m.fs.cv, "reactions")
    assert len(m.fs.cv.reactions[0].config) == 3
    assert m.fs.cv.reactions[0].config.state_block == m.fs.cv.properties_out
    assert m.fs.cv.reactions[0].state_ref == m.fs.cv.properties_out[0]
    assert m.fs.cv.reactions[0].config.has_equilibrium is False
    assert m.fs.cv.reactions[0].config.parameters == m.fs.rp


@pytest.mark.unit
def test_add_reaction_blocks_has_equilibrium():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=True)

    assert m.fs.cv.reactions[0].config.has_equilibrium is True


@pytest.mark.unit
def test_add_reaction_blocks_no_has_equilibrium():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_reaction_blocks()


@pytest.mark.unit
def test_add_reaction_blocks_custom_args():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(
        property_package=m.fs.pp,
        reaction_package=m.fs.rp,
        reaction_package_args={"test1": 1},
    )

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=True)

    assert m.fs.cv.reactions[0].config.test1 == 1


# -----------------------------------------------------------------------------
# Test _add_phase_fractions
@pytest.mark.unit
def test_add_phase_fractions():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv._add_phase_fractions()

    assert isinstance(m.fs.cv.phase_fraction, Var)
    assert len(m.fs.cv.phase_fraction) == 2
    assert hasattr(m.fs.cv, "sum_of_phase_fractions")


@pytest.mark.unit
def test_add_phase_fractions_single_phase():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.del_component(m.fs.pp.phase_list)
    m.fs.pp.phase_list = Set(initialize=["p1"])

    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv._add_phase_fractions()

    assert isinstance(m.fs.cv.phase_fraction, Expression)
    assert len(m.fs.cv.phase_fraction) == 1
    assert not hasattr(m.fs.cv, "sum_of_phase_fractions")


# -----------------------------------------------------------------------------
# Test reaction rate conversion method
@pytest.mark.unit
def test_rxn_rate_conv_property_basis_other():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.basis_switch = 3
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    for t in m.fs.time:
        for j in m.fs.pp.component_list:
            with pytest.raises(ConfigurationError):
                m.fs.cv._rxn_rate_conv(t, j)


@pytest.mark.unit
def test_rxn_rate_conv_reaction_basis_other():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.rp.basis_switch = 3

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    for t in m.fs.time:
        for j in m.fs.pp.component_list:
            with pytest.raises(ConfigurationError):
                m.fs.cv._rxn_rate_conv(t, j)


@pytest.mark.unit
def test_rxn_rate_conv_both_molar():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    for t in m.fs.time:
        for j in m.fs.pp.component_list:
            assert m.fs.cv._rxn_rate_conv(t, j) == 1


@pytest.mark.unit
def test_rxn_rate_conv_both_mass():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.basis_switch = 2
    m.fs.rp.basis_switch = 2

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    for t in m.fs.time:
        for j in m.fs.pp.component_list:
            assert m.fs.cv._rxn_rate_conv(t, j) == 1


@pytest.mark.unit
def test_rxn_rate_conv_mole_mass_no_mw():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.basis_switch = 1
    m.fs.rp.basis_switch = 2

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    for t in m.fs.time:
        for j in m.fs.pp.component_list:
            with pytest.raises(PropertyNotSupportedError):
                m.fs.cv._rxn_rate_conv(t, j)


@pytest.mark.unit
def test_rxn_rate_conv_mass_mole_no_mw():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.basis_switch = 2
    m.fs.rp.basis_switch = 1

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    for t in m.fs.time:
        for j in m.fs.pp.component_list:
            with pytest.raises(PropertyNotSupportedError):
                m.fs.cv._rxn_rate_conv(t, j)


@pytest.mark.unit
def test_rxn_rate_conv_mole_mass():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.basis_switch = 1
    m.fs.rp.basis_switch = 2

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    for t in m.fs.time:
        m.fs.cv.properties_out[t].mw_comp = {"c1": 2, "c2": 3}
        for j in m.fs.pp.component_list:
            assert (
                m.fs.cv._rxn_rate_conv(t, j) == 1 / m.fs.cv.properties_out[t].mw_comp[j]
            )


@pytest.mark.unit
def test_rxn_rate_conv_mass_mole():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.basis_switch = 2
    m.fs.rp.basis_switch = 1

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    for t in m.fs.time:
        m.fs.cv.properties_out[t].mw_comp = {"c1": 2, "c2": 3}
        for j in m.fs.pp.component_list:
            assert m.fs.cv._rxn_rate_conv(t, j) == m.fs.cv.properties_out[t].mw_comp[j]


# -----------------------------------------------------------------------------
# Test add_material_balances default
@pytest.mark.unit
def test_add_material_balances_default_fail():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.pp.default_balance_switch = 2

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_material_balances(MaterialBalanceType.useDefault)


@pytest.mark.unit
def test_add_material_balances_default():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_material_balances(MaterialBalanceType.useDefault)

    assert isinstance(mb, Constraint)
    assert len(mb) == 4

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_material_balances_rxn_molar():
    # use property package with mass basis to confirm correct rxn term units
    # add options so that all generation/extent terms exist
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()

    # Set property package to contain inherent reactions
    m.fs.pp._has_inherent_reactions = True

    m.fs.pp.basis_switch = 2
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.rp.basis_switch = 1

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    units = m.fs.cv.config.property_package.get_metadata().get_derived_units
    pp_units = units("flow_mass")  # basis 2 is mass
    rp_units = units("flow_mole")  # basis 1 is molar

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=True)

    # add molecular weight variable to each time point, using correct units
    for t in m.fs.time:
        m.fs.cv.properties_out[t].mw_comp = Var(
            m.fs.cv.properties_out[t].config.parameters.component_list,
            units=units("mass") / units("amount"),
        )

    # add material balances to control volume
    m.fs.cv.add_material_balances(
        balance_type=MaterialBalanceType.componentPhase,
        has_rate_reactions=True,
        has_equilibrium_reactions=True,
        has_phase_equilibrium=True,
    )

    assert_units_equivalent(m.fs.cv.rate_reaction_generation, rp_units)
    assert_units_equivalent(m.fs.cv.rate_reaction_extent, rp_units)
    assert_units_equivalent(m.fs.cv.equilibrium_reaction_generation, rp_units)
    assert_units_equivalent(m.fs.cv.equilibrium_reaction_extent, rp_units)
    assert_units_equivalent(m.fs.cv.inherent_reaction_generation, pp_units)
    assert_units_equivalent(m.fs.cv.inherent_reaction_extent, pp_units)
    assert_units_equivalent(m.fs.cv.phase_equilibrium_generation, pp_units)


@pytest.mark.unit
def test_add_material_balances_rxn_mass():
    # use property package with mass basis to confirm correct rxn term units
    # add options so that all generation/extent terms exist
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()

    # Set property package to contain inherent reactions
    m.fs.pp._has_inherent_reactions = True

    m.fs.pp.basis_switch = 1
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.rp.basis_switch = 2

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    units = m.fs.cv.config.property_package.get_metadata().get_derived_units
    pp_units = units("flow_mole")  # basis 2 is molar
    rp_units = units("flow_mass")  # basis 1 is mass

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=True)

    # add molecular weight variable to each time point, using correct units
    for t in m.fs.time:
        m.fs.cv.properties_out[t].mw_comp = Var(
            m.fs.cv.properties_out[t].config.parameters.component_list,
            units=units("mass") / units("amount"),
        )

    # add material balances to control volume
    m.fs.cv.add_material_balances(
        balance_type=MaterialBalanceType.componentPhase,
        has_rate_reactions=True,
        has_equilibrium_reactions=True,
        has_phase_equilibrium=True,
    )

    assert_units_equivalent(m.fs.cv.rate_reaction_generation, rp_units)
    assert_units_equivalent(m.fs.cv.rate_reaction_extent, rp_units)
    assert_units_equivalent(m.fs.cv.equilibrium_reaction_generation, rp_units)
    assert_units_equivalent(m.fs.cv.equilibrium_reaction_extent, rp_units)
    assert_units_equivalent(m.fs.cv.inherent_reaction_generation, pp_units)
    assert_units_equivalent(m.fs.cv.inherent_reaction_extent, pp_units)
    assert_units_equivalent(m.fs.cv.phase_equilibrium_generation, pp_units)


@pytest.mark.unit
def test_add_material_balances_single_phase_w_equilibrium(caplog):
    from idaes.models.properties import iapws95

    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = iapws95.Iapws95ParameterBlock(
        phase_presentation=iapws95.PhaseType.MIX, state_vars=iapws95.StateVars.PH
    )

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)

    m.fs.cv.add_material_balances(
        balance_type=MaterialBalanceType.useDefault,
        has_phase_equilibrium=True,
    )
    msg = (
        "DEPRECATED: Property package has only one phase; control volume cannot "
        "include phase equilibrium terms. Some property packages support phase "
        "equilibrium implicitly in which case additional terms are not "
        "necessary. You should set has_phase_equilibrium=False.  (deprecated in "
        "2.0.0, will be removed in (or after) 3.0.0)"
    )
    assert msg.replace(" ", "") in caplog.records[0].message.replace("\n", "").replace(
        " ", ""
    )


# -----------------------------------------------------------------------------
# Test add_phase_component_balances
@pytest.mark.unit
def test_add_phase_component_balances_default():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_phase_component_balances()

    assert isinstance(mb, Constraint)
    assert len(mb) == 4

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_phase_component_balances_dynamic():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=True, time_units=units.s)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(
        property_package=m.fs.pp, reaction_package=m.fs.rp, dynamic=True
    )

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_phase_component_balances()

    assert isinstance(mb, Constraint)
    assert len(mb) == 8
    assert isinstance(m.fs.cv.phase_fraction, Var)
    assert isinstance(m.fs.cv.material_holdup, Var)
    assert isinstance(m.fs.cv.material_accumulation, Var)

    assert_units_consistent(m)
    assert_units_equivalent(m.fs.cv.material_holdup, units.mol)
    assert_units_equivalent(m.fs.cv.material_accumulation, units.mol / units.s)


@pytest.mark.unit
def test_add_phase_component_balances_dynamic_no_geometry():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=True, time_units=units.s)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(
        property_package=m.fs.pp, reaction_package=m.fs.rp, dynamic=True
    )

    # Do not add geometry
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_phase_component_balances()


@pytest.mark.unit
def test_add_phase_component_balances_rate_rxns():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_phase_component_balances(has_rate_reactions=True)

    assert isinstance(mb, Constraint)
    assert len(mb) == 4
    assert isinstance(m.fs.cv.rate_reaction_generation, Var)
    assert isinstance(m.fs.cv.rate_reaction_extent, Var)
    assert isinstance(m.fs.cv.rate_reaction_stoichiometry_constraint, Constraint)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_phase_component_balances_rate_rxns_no_ReactionBlock():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_phase_component_balances(has_rate_reactions=True)


@pytest.mark.unit
def test_add_phase_component_balances_rate_rxns_no_rxn_idx():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.rp.del_component(m.fs.rp.rate_reaction_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(PropertyNotSupportedError):
        m.fs.cv.add_phase_component_balances(has_rate_reactions=True)


@pytest.mark.unit
def test_add_phase_component_balances_eq_rxns():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=True)

    mb = m.fs.cv.add_phase_component_balances(has_equilibrium_reactions=True)

    assert isinstance(mb, Constraint)
    assert len(mb) == 4
    assert isinstance(m.fs.cv.equilibrium_reaction_generation, Var)
    assert isinstance(m.fs.cv.equilibrium_reaction_extent, Var)
    assert isinstance(m.fs.cv.equilibrium_reaction_stoichiometry_constraint, Constraint)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_phase_component_balances_eq_rxns_not_active():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_phase_component_balances(has_equilibrium_reactions=True)


@pytest.mark.unit
def test_add_phase_component_balances_eq_rxns_no_idx():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.rp.del_component(m.fs.rp.equilibrium_reaction_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=True)

    with pytest.raises(PropertyNotSupportedError):
        m.fs.cv.add_phase_component_balances(has_equilibrium_reactions=True)


@pytest.mark.unit
def test_add_phase_component_balances_in_rxns():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()

    # Set property package to contain inherent reactions
    m.fs.pp._has_inherent_reactions = True

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)

    mb = m.fs.cv.add_phase_component_balances()

    assert isinstance(mb, Constraint)
    assert len(mb) == 4
    assert isinstance(m.fs.cv.inherent_reaction_generation, Var)
    assert isinstance(m.fs.cv.inherent_reaction_extent, Var)
    assert isinstance(m.fs.cv.inherent_reaction_stoichiometry_constraint, Constraint)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_phase_component_balances_in_rxns_no_idx():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()

    # Set property package to contain inherent reactions
    m.fs.pp._has_inherent_reactions = True
    # delete inherent_Reaction_dix to trigger exception
    m.fs.pp.del_component(m.fs.pp.inherent_reaction_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)

    with pytest.raises(
        PropertyNotSupportedError,
        match=r"fs.cv Property package does not contain a "
        r"list of inherent reactions \(inherent_reaction_idx\), "
        r"but include_inherent_reactions is True.",
    ):
        m.fs.cv.add_phase_component_balances()


@pytest.mark.unit
def test_add_phase_component_balances_eq_rxns_no_ReactionBlock():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_phase_component_balances(has_equilibrium_reactions=True)


@pytest.mark.unit
def test_add_phase_component_balances_phase_eq():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_phase_component_balances(has_phase_equilibrium=True)

    assert isinstance(mb, Constraint)
    assert len(mb) == 4
    assert isinstance(m.fs.cv.phase_equilibrium_generation, Var)
    assert isinstance(m.fs.cv.config.property_package.phase_equilibrium_idx, Set)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_phase_component_balances_phase_eq_not_active():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_phase_component_balances(has_phase_equilibrium=True)


@pytest.mark.unit
def test_add_phase_component_balances_phase_eq_no_idx():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.del_component(m.fs.pp.phase_equilibrium_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(PropertyNotSupportedError):
        m.fs.cv.add_phase_component_balances(has_phase_equilibrium=True)


@pytest.mark.unit
def test_add_phase_component_balances_mass_transfer():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_phase_component_balances(has_mass_transfer=True)

    assert isinstance(mb, Constraint)
    assert len(mb) == 4
    assert isinstance(m.fs.cv.mass_transfer_term, Var)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_phase_component_balances_custom_molar_term():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time,
        m.fs.cv.config.property_package.phase_list,
        m.fs.cv.config.property_package.component_list,
    )

    def custom_method(t, p, j):
        return m.fs.cv.test_var[t, p, j] * units.mol / units.s

    mb = m.fs.cv.add_phase_component_balances(custom_molar_term=custom_method)

    assert isinstance(mb, Constraint)
    assert len(mb) == 4

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_phase_component_balances_custom_molar_term_no_mw():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.basis_switch = 2
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time,
        m.fs.cv.config.property_package.phase_list,
        m.fs.cv.config.property_package.component_list,
    )

    def custom_method(t, p, j):
        return m.fs.cv.test_var[t, p, j]

    with pytest.raises(PropertyNotSupportedError):
        m.fs.cv.add_phase_component_balances(custom_molar_term=custom_method)


@pytest.mark.unit
def test_add_phase_component_balances_custom_molar_term_mass_flow_basis():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.basis_switch = 2
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time,
        m.fs.cv.config.property_package.phase_list,
        m.fs.cv.config.property_package.component_list,
    )

    def custom_method(t, p, j):
        return m.fs.cv.test_var[t, p, j] * units.mol / units.s

    for t in m.fs.time:
        m.fs.cv.properties_out[t].mw_comp = Var(
            m.fs.cv.properties_out[t].config.parameters.component_list,
            units=units.kg / units.mol,
        )

    mb = m.fs.cv.add_phase_component_balances(custom_molar_term=custom_method)

    assert isinstance(mb, Constraint)
    assert len(mb) == 4

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_phase_component_balances_custom_molar_term_undefined_basis():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.basis_switch = 3
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time,
        m.fs.cv.config.property_package.phase_list,
        m.fs.cv.config.property_package.component_list,
    )

    def custom_method(t, p, j):
        return m.fs.cv.test_var[t, p, j]

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_phase_component_balances(custom_molar_term=custom_method)


@pytest.mark.unit
def test_add_phase_component_balances_custom_mass_term():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.basis_switch = 2
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time,
        m.fs.cv.config.property_package.phase_list,
        m.fs.cv.config.property_package.component_list,
    )

    def custom_method(t, p, j):
        return m.fs.cv.test_var[t, p, j] * units.kg / units.s

    mb = m.fs.cv.add_phase_component_balances(custom_mass_term=custom_method)

    assert isinstance(mb, Constraint)
    assert len(mb) == 4

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_phase_component_balances_custom_mass_term_no_mw():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.basis_switch = 1
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time,
        m.fs.cv.config.property_package.phase_list,
        m.fs.cv.config.property_package.component_list,
    )

    def custom_method(t, p, j):
        return m.fs.cv.test_var[t, p, j]

    with pytest.raises(PropertyNotSupportedError):
        m.fs.cv.add_phase_component_balances(custom_mass_term=custom_method)


@pytest.mark.unit
def test_add_phase_component_balances_custom_mass_term_mole_flow_basis():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.basis_switch = 2
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time,
        m.fs.cv.config.property_package.phase_list,
        m.fs.cv.config.property_package.component_list,
    )

    def custom_method(t, p, j):
        return m.fs.cv.test_var[t, p, j] * units.kg / units.s

    for t in m.fs.time:
        m.fs.cv.properties_out[t].mw_comp = Var(
            m.fs.cv.properties_out[t].config.parameters.component_list,
            units=units.kg / units.mol,
        )

    mb = m.fs.cv.add_phase_component_balances(custom_mass_term=custom_method)

    assert isinstance(mb, Constraint)
    assert len(mb) == 4

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_phase_component_balances_custom_mass_term_undefined_basis():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.basis_switch = 3
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time,
        m.fs.cv.config.property_package.phase_list,
        m.fs.cv.config.property_package.component_list,
    )

    def custom_method(t, p, j):
        return m.fs.cv.test_var[t, p, j]

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_phase_component_balances(custom_mass_term=custom_method)


# -----------------------------------------------------------------------------
# Test add_total_component_balances
@pytest.mark.unit
def test_add_total_component_balances_default():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_total_component_balances()

    assert isinstance(mb, Constraint)
    assert len(mb) == 2

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_component_balances_dynamic():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=True, time_units=units.s)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(
        property_package=m.fs.pp, reaction_package=m.fs.rp, dynamic=True
    )

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_total_component_balances()

    assert isinstance(mb, Constraint)
    assert len(mb) == 4
    assert isinstance(m.fs.cv.phase_fraction, Var)
    assert isinstance(m.fs.cv.material_holdup, Var)
    assert isinstance(m.fs.cv.material_accumulation, Var)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_component_balances_dynamic_no_geometry():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=True, time_units=units.s)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(
        property_package=m.fs.pp, reaction_package=m.fs.rp, dynamic=True
    )

    # Do not add geometry
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_total_component_balances()


@pytest.mark.unit
def test_add_total_component_balances_rate_rxns():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_total_component_balances(has_rate_reactions=True)

    assert isinstance(mb, Constraint)
    assert len(mb) == 2
    assert isinstance(m.fs.cv.rate_reaction_generation, Var)
    assert isinstance(m.fs.cv.rate_reaction_extent, Var)
    assert isinstance(m.fs.cv.rate_reaction_stoichiometry_constraint, Constraint)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_component_balances_rate_rxns_no_ReactionBlock():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_total_component_balances(has_rate_reactions=True)


@pytest.mark.unit
def test_add_total_component_balances_rate_rxns_no_rxn_idx():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.rp.del_component(m.fs.rp.rate_reaction_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(PropertyNotSupportedError):
        m.fs.cv.add_total_component_balances(has_rate_reactions=True)


@pytest.mark.unit
def test_add_total_component_balances_eq_rxns():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=True)

    mb = m.fs.cv.add_total_component_balances(has_equilibrium_reactions=True)

    assert isinstance(mb, Constraint)
    assert len(mb) == 2
    assert isinstance(m.fs.cv.equilibrium_reaction_generation, Var)
    assert isinstance(m.fs.cv.equilibrium_reaction_extent, Var)
    assert isinstance(m.fs.cv.equilibrium_reaction_stoichiometry_constraint, Constraint)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_component_balances_eq_rxns_not_active():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_total_component_balances(has_equilibrium_reactions=True)


@pytest.mark.unit
def test_add_total_component_balances_eq_rxns_no_idx():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.rp.del_component(m.fs.rp.equilibrium_reaction_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=True)

    with pytest.raises(PropertyNotSupportedError):
        m.fs.cv.add_total_component_balances(has_equilibrium_reactions=True)


@pytest.mark.unit
def test_add_total_component_balances_eq_rxns_no_ReactionBlock():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_total_component_balances(has_equilibrium_reactions=True)


@pytest.mark.unit
def test_add_total_component_balances_in_rxns():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()

    # Set property package to contain inherent reactions
    m.fs.pp._has_inherent_reactions = True

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)

    mb = m.fs.cv.add_total_component_balances()

    assert isinstance(mb, Constraint)
    assert len(mb) == 2
    assert isinstance(m.fs.cv.inherent_reaction_generation, Var)
    assert isinstance(m.fs.cv.inherent_reaction_extent, Var)
    assert isinstance(m.fs.cv.inherent_reaction_stoichiometry_constraint, Constraint)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_component_balances_in_rxns_no_idx():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()

    # Set property package to contain inherent reactions
    m.fs.pp._has_inherent_reactions = True
    # delete inherent_Reaction_dix to trigger exception
    m.fs.pp.del_component(m.fs.pp.inherent_reaction_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)

    with pytest.raises(
        PropertyNotSupportedError,
        match=r"fs.cv Property package does not contain a "
        r"list of inherent reactions \(inherent_reaction_idx\), "
        r"but include_inherent_reactions is True.",
    ):
        m.fs.cv.add_total_component_balances()


@pytest.mark.unit
def test_add_total_component_balances_phase_eq_not_active():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_total_component_balances(has_phase_equilibrium=True)


@pytest.mark.unit
def test_add_total_component_balances_mass_transfer():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_total_component_balances(has_mass_transfer=True)

    assert isinstance(mb, Constraint)
    assert len(mb) == 2
    assert isinstance(m.fs.cv.mass_transfer_term, Var)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_component_balances_custom_molar_term():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time, m.fs.cv.config.property_package.component_list
    )

    def custom_method(t, j):
        return m.fs.cv.test_var[t, j] * units.mol / units.s

    mb = m.fs.cv.add_total_component_balances(custom_molar_term=custom_method)

    assert isinstance(mb, Constraint)
    assert len(mb) == 2

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_component_balances_custom_molar_term_no_mw():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.basis_switch = 2
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time, m.fs.cv.config.property_package.component_list
    )

    def custom_method(t, j):
        return m.fs.cv.test_var[t, j]

    with pytest.raises(PropertyNotSupportedError):
        m.fs.cv.add_total_component_balances(custom_molar_term=custom_method)


@pytest.mark.unit
def test_add_total_component_balances_custom_molar_term_mass_flow_basis():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.basis_switch = 2
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time, m.fs.cv.config.property_package.component_list
    )

    def custom_method(t, j):
        return m.fs.cv.test_var[t, j] * units.mol / units.s

    for t in m.fs.time:
        m.fs.cv.properties_out[t].mw_comp = Var(
            m.fs.cv.properties_out[t].config.parameters.component_list,
            units=units.kg / units.mol,
        )

    mb = m.fs.cv.add_total_component_balances(custom_molar_term=custom_method)

    assert isinstance(mb, Constraint)
    assert len(mb) == 2

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_component_balances_custom_molar_term_undefined_basis():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.basis_switch = 3
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time, m.fs.cv.config.property_package.component_list
    )

    def custom_method(t, j):
        return m.fs.cv.test_var[t, j]

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_total_component_balances(custom_molar_term=custom_method)


@pytest.mark.unit
def test_add_total_component_balances_custom_mass_term():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.basis_switch = 2
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time, m.fs.cv.config.property_package.component_list
    )

    def custom_method(t, j):
        return m.fs.cv.test_var[t, j] * units.kg / units.s

    mb = m.fs.cv.add_total_component_balances(custom_mass_term=custom_method)

    assert isinstance(mb, Constraint)
    assert len(mb) == 2

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_component_balances_custom_mass_term_no_mw():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.basis_switch = 1
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time, m.fs.cv.config.property_package.component_list
    )

    def custom_method(t, j):
        return m.fs.cv.test_var[t, j]

    with pytest.raises(PropertyNotSupportedError):
        m.fs.cv.add_total_component_balances(custom_mass_term=custom_method)


@pytest.mark.unit
def test_add_total_component_balances_custom_mass_term_mole_flow_basis():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.basis_switch = 2
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time, m.fs.cv.config.property_package.component_list
    )

    def custom_method(t, j):
        return m.fs.cv.test_var[t, j] * units.kg / units.s

    for t in m.fs.time:
        m.fs.cv.properties_out[t].mw_comp = Var(
            m.fs.cv.properties_out[t].config.parameters.component_list,
            units=units.kg / units.mol,
        )

    mb = m.fs.cv.add_total_component_balances(custom_mass_term=custom_method)

    assert isinstance(mb, Constraint)
    assert len(mb) == 2

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_component_balances_custom_mass_term_undefined_basis():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.pp.basis_switch = 3
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(
        m.fs.cv.flowsheet().time, m.fs.cv.config.property_package.component_list
    )

    def custom_method(t, j):
        return m.fs.cv.test_var[t, j]

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_total_component_balances(custom_mass_term=custom_method)


# -----------------------------------------------------------------------------
# Test add_total_element_balances
@pytest.mark.unit
def test_add_total_element_balances_default():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_total_element_balances()

    assert isinstance(mb, Constraint)
    assert len(mb) == 3

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_element_balances_properties_not_supported():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.del_component(m.fs.pp.element_list)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(PropertyNotSupportedError):
        m.fs.cv.add_total_element_balances()


@pytest.mark.unit
def test_add_total_element_balances_dynamic():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=True, time_units=units.s)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(
        property_package=m.fs.pp, reaction_package=m.fs.rp, dynamic=True
    )

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_total_element_balances()

    assert isinstance(mb, Constraint)
    assert len(mb) == 6
    assert isinstance(m.fs.cv.phase_fraction, Var)
    assert isinstance(m.fs.cv.element_holdup, Var)
    assert isinstance(m.fs.cv.element_accumulation, Var)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_element_balances_dynamic_no_geometry():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=True, time_units=units.s)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(
        property_package=m.fs.pp, reaction_package=m.fs.rp, dynamic=True
    )

    # Do not add geometry
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_total_element_balances()


@pytest.mark.unit
def test_add_total_element_balances_rate_rxns():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_total_element_balances(has_rate_reactions=True)


@pytest.mark.unit
def test_add_total_element_balances_eq_rxns():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=True)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_total_element_balances(has_equilibrium_reactions=True)


@pytest.mark.unit
def test_add_total_element_balances_phase_eq():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_total_element_balances(has_phase_equilibrium=True)


@pytest.mark.unit
def test_add_total_element_balances_mass_transfer():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_total_element_balances(has_mass_transfer=True)

    assert isinstance(mb, Constraint)
    assert len(mb) == 3
    assert isinstance(m.fs.cv.elemental_mass_transfer_term, Var)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_element_balances_custom_term():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(m.fs.cv.flowsheet().time, m.fs.pp.element_list)

    def custom_method(t, e):
        return m.fs.cv.test_var[t, e] * units.mol / units.s

    mb = m.fs.cv.add_total_element_balances(custom_elemental_term=custom_method)

    assert isinstance(mb, Constraint)
    assert len(mb) == 3

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_element_balances_lineraly_dependent(caplog):
    caplog.set_level(idaeslog.INFO_LOW)
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()

    # Change elemental composition to introduce dependency
    m.fs.pp.element_comp = {
        "c1": {"H": 0, "He": 0, "Li": 1},
        "c2": {"H": 1, "He": 2, "Li": 0},
    }

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)

    mb = m.fs.cv.add_total_element_balances()
    # Check that logger message was recorded and has the right level
    msg = (
        "fs.cv detected linearly dependent element balance equations. "
        "Element balances will NOT be written for the following elements: "
        "['He']"
    )
    assert msg in caplog.text
    for record in caplog.records:
        if "['He']" in record.msg:
            assert record.levelno == idaeslog.INFO_LOW

    assert isinstance(mb, Constraint)
    assert len(mb) == 2
    for e in mb:
        # H and Li are not lineraly dependent and should have constraints
        assert e in [(0, "H"), (0, "Li")]
        # He is lineraly dependent on H and should be skipped

    assert_units_consistent(m)


# -----------------------------------------------------------------------------
# Test unsupported material balance types
@pytest.mark.unit
def test_add_total_material_balances():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.del_component(m.fs.pp.phase_equilibrium_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(BalanceTypeNotSupportedError):
        m.fs.cv.add_total_material_balances()


# -----------------------------------------------------------------------------
# Test add_energy_balances default
@pytest.mark.unit
def test_add_energy_balances_default_fail():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.pp.default_balance_switch = 2

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_energy_balances(EnergyBalanceType.useDefault)


@pytest.mark.unit
def test_add_energy_balances_default():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    eb = m.fs.cv.add_energy_balances(EnergyBalanceType.useDefault)

    assert isinstance(eb, Constraint)
    assert len(eb) == 1

    assert_units_consistent(m)


# -----------------------------------------------------------------------------
# Test phase enthalpy balances
@pytest.mark.unit
def test_add_total_enthalpy_balances_default():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    eb = m.fs.cv.add_total_enthalpy_balances()

    assert isinstance(eb, Constraint)
    assert len(eb) == 1

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_enthalpy_balances_dynamic():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=True, time_units=units.s)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(
        property_package=m.fs.pp, reaction_package=m.fs.rp, dynamic=True
    )

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_total_enthalpy_balances()

    assert isinstance(mb, Constraint)
    assert len(mb) == 2
    assert isinstance(m.fs.cv.phase_fraction, Var)
    assert isinstance(m.fs.cv.energy_holdup, Var)
    assert isinstance(m.fs.cv.energy_accumulation, Var)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_enthalpy_balances_dynamic_no_geometry():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=True, time_units=units.s)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(
        property_package=m.fs.pp, reaction_package=m.fs.rp, dynamic=True
    )

    # Do not add geometry
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_total_enthalpy_balances()


@pytest.mark.unit
def test_add_total_enthalpy_balances_heat_transfer():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_total_enthalpy_balances(has_heat_transfer=True)

    assert isinstance(mb, Constraint)
    assert len(mb) == 1
    assert isinstance(m.fs.cv.heat, Var)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_enthalpy_balances_work_transfer():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_total_enthalpy_balances(has_work_transfer=True)

    assert isinstance(mb, Constraint)
    assert len(mb) == 1
    assert isinstance(m.fs.cv.work, Var)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_enthalpy_balances_enthalpy_transfer():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_total_enthalpy_balances(has_enthalpy_transfer=True)

    assert isinstance(mb, Constraint)
    assert len(mb) == 1
    assert isinstance(m.fs.cv.enthalpy_transfer, Var)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_enthalpy_balances_custom_term():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(m.fs.cv.flowsheet().time)

    def custom_method(t):
        return m.fs.cv.test_var[t] * units.J / units.s

    mb = m.fs.cv.add_total_enthalpy_balances(custom_term=custom_method)

    assert isinstance(mb, Constraint)
    assert len(mb) == 1

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_enthalpy_balances_dh_rxn_no_extents():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.cv.add_total_enthalpy_balances(has_heat_of_reaction=True)


@pytest.mark.unit
def test_add_total_enthalpy_balances_dh_rxn_rate_rxns():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)
    m.fs.cv.add_phase_component_balances(has_rate_reactions=True)

    m.fs.cv.add_total_enthalpy_balances(has_heat_of_reaction=True)
    assert isinstance(m.fs.cv.heat_of_reaction, Expression)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_enthalpy_balances_dh_rxn_equil_rxns():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=True)
    m.fs.cv.add_phase_component_balances(has_equilibrium_reactions=True)

    m.fs.cv.add_total_enthalpy_balances(has_heat_of_reaction=True)
    assert isinstance(m.fs.cv.heat_of_reaction, Expression)

    assert_units_consistent(m)


# -----------------------------------------------------------------------------
# Test unsupported energy balance types
@pytest.mark.unit
def test_add_phase_enthalpy_balances():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.del_component(m.fs.pp.phase_equilibrium_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(BalanceTypeNotSupportedError):
        m.fs.cv.add_phase_enthalpy_balances()


@pytest.mark.unit
def test_add_phase_energy_balances():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.del_component(m.fs.pp.phase_equilibrium_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(BalanceTypeNotSupportedError):
        m.fs.cv.add_phase_energy_balances()


@pytest.mark.unit
def test_add_total_energy_balances():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.del_component(m.fs.pp.phase_equilibrium_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(BalanceTypeNotSupportedError):
        m.fs.cv.add_total_energy_balances()


# -----------------------------------------------------------------------------
# Test add total pressure balances
@pytest.mark.unit
def test_add_total_pressure_balances_default():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    eb = m.fs.cv.add_total_pressure_balances()

    assert isinstance(eb, Constraint)
    assert len(eb) == 1

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_pressure_balances_deltaP():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_total_pressure_balances(has_pressure_change=True)

    assert isinstance(mb, Constraint)
    assert len(mb) == 1
    assert isinstance(m.fs.cv.deltaP, Var)

    assert_units_consistent(m)


@pytest.mark.unit
def test_add_total_pressure_balances_custom_term():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.test_var = Var(m.fs.cv.flowsheet().time)

    def custom_method(t):
        return m.fs.cv.test_var[t] * units.Pa

    mb = m.fs.cv.add_total_pressure_balances(custom_term=custom_method)

    assert isinstance(mb, Constraint)
    assert len(mb) == 1

    assert_units_consistent(m)


# -----------------------------------------------------------------------------
# Test unsupported momentum balance types
@pytest.mark.unit
def test_add_phase_pressure_balances():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.del_component(m.fs.pp.phase_equilibrium_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(BalanceTypeNotSupportedError):
        m.fs.cv.add_phase_pressure_balances()


@pytest.mark.unit
def test_add_phase_momentum_balances():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.del_component(m.fs.pp.phase_equilibrium_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(BalanceTypeNotSupportedError):
        m.fs.cv.add_phase_momentum_balances()


@pytest.mark.unit
def test_add_total_momentum_balances():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.del_component(m.fs.pp.phase_equilibrium_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    with pytest.raises(BalanceTypeNotSupportedError):
        m.fs.cv.add_total_momentum_balances()


# -----------------------------------------------------------------------------
# Test model checks, initialize and release_state
@pytest.mark.unit
def test_model_checks():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.del_component(m.fs.pp.phase_equilibrium_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    m.fs.cv.model_check()

    for t in m.fs.time:
        assert m.fs.cv.properties_in[t].check is True
        assert m.fs.cv.properties_out[t].check is True
        assert m.fs.cv.reactions[t].check is True


@pytest.mark.unit
def test_initialize():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.del_component(m.fs.pp.phase_equilibrium_idx)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    f = m.fs.cv.initialize()

    for t in m.fs.time:
        assert m.fs.cv.properties_in[t].init_test is True
        assert m.fs.cv.properties_out[t].init_test is True
        assert m.fs.cv.properties_in[t].hold_state is True
        assert m.fs.cv.properties_out[t].hold_state is False
        assert m.fs.cv.reactions[t].init_test is True

    m.fs.cv.release_state(flags=f)

    for t in m.fs.time:
        assert m.fs.cv.properties_in[t].hold_state is False
        assert m.fs.cv.properties_out[t].hold_state is False


@pytest.mark.unit
def test_get_stream_table_contents():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)

    df = m.fs.cv._get_stream_table_contents()

    assert df.loc["component_flow_phase ('p1', 'c1')"]["In"] == 2
    assert df.loc["component_flow_phase ('p1', 'c2')"]["In"] == 2
    assert df.loc["component_flow_phase ('p2', 'c1')"]["In"] == 2
    assert df.loc["component_flow_phase ('p2', 'c2')"]["In"] == 2
    assert df.loc["pressure"]["In"] == 1e5
    assert df.loc["temperature"]["In"] == 300

    assert df.loc["component_flow_phase ('p1', 'c1')"]["Out"] == 2
    assert df.loc["component_flow_phase ('p1', 'c2')"]["Out"] == 2
    assert df.loc["component_flow_phase ('p2', 'c1')"]["Out"] == 2
    assert df.loc["component_flow_phase ('p2', 'c2')"]["Out"] == 2
    assert df.loc["pressure"]["Out"] == 1e5
    assert df.loc["temperature"]["Out"] == 300


@pytest.mark.unit
def test_get_performance_contents():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=True, time_units=units.s)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=True)
    m.fs.cv.add_material_balances(
        has_rate_reactions=True,
        has_equilibrium_reactions=True,
        has_phase_equilibrium=True,
        has_mass_transfer=True,
    )
    m.fs.cv.add_energy_balances(
        has_heat_of_reaction=True, has_work_transfer=True, has_heat_transfer=True
    )
    m.fs.cv.add_momentum_balances(has_pressure_change=True)

    dd = m.fs.cv._get_performance_contents()

    assert len(dd) == 3
    for k in dd.keys():
        assert k in ("vars", "exprs", "params")
    assert len(dd["vars"]) == 36
    for k in dd["vars"].keys():
        assert k in [
            "Volume",
            "Heat Transfer",
            "Work Transfer",
            "Pressure Change",
            "Phase Fraction [p1]",
            "Phase Fraction [p2]",
            "Energy Holdup [p1]",
            "Energy Holdup [p2]",
            "Energy Accumulation [p1]",
            "Energy Accumulation [p2]",
            "Material Holdup [p1, c1]",
            "Material Holdup [p1, c2]",
            "Material Holdup [p2, c1]",
            "Material Holdup [p2, c2]",
            "Material Accumulation [p1, c1]",
            "Material Accumulation [p1, c2]",
            "Material Accumulation [p2, c1]",
            "Material Accumulation [p2, c2]",
            "Rate Reaction Generation [p1, c1]",
            "Rate Reaction Generation [p1, c2]",
            "Rate Reaction Generation [p2, c1]",
            "Rate Reaction Generation [p2, c2]",
            "Equilibrium Reaction Generation [p1, c1]",
            "Equilibrium Reaction Generation [p1, c2]",
            "Equilibrium Reaction Generation [p2, c1]",
            "Equilibrium Reaction Generation [p2, c2]",
            "Mass Transfer Term [p1, c1]",
            "Mass Transfer Term [p1, c2]",
            "Mass Transfer Term [p2, c1]",
            "Mass Transfer Term [p2, c2]",
            "Rate Reaction Extent [r1]",
            "Rate Reaction Extent [r2]",
            "Equilibrium Reaction Extent [e1]",
            "Equilibrium Reaction Extent [e2]",
            "Phase Equilibrium Generation [e1]",
            "Phase Equilibrium Generation [e2]",
        ]

    assert len(dd["exprs"]) == 1
    for k in dd["exprs"].keys():
        assert k in ["Heat of Reaction Term"]

    assert len(dd["params"]) == 0


@pytest.mark.unit
def test_get_performance_contents_elemental():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=True, time_units=units.s)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=True)
    m.fs.cv.add_total_element_balances(has_mass_transfer=True)
    m.fs.cv.add_energy_balances(
        has_heat_of_reaction=False, has_work_transfer=True, has_heat_transfer=True
    )
    m.fs.cv.add_momentum_balances(has_pressure_change=True)

    dd = m.fs.cv._get_performance_contents()

    assert len(dd) == 3
    for k in dd.keys():
        assert k in ("vars", "exprs", "params")
    assert len(dd["vars"]) == 19
    for k in dd["vars"].keys():
        assert k in [
            "Volume",
            "Heat Transfer",
            "Work Transfer",
            "Pressure Change",
            "Phase Fraction [p1]",
            "Phase Fraction [p2]",
            "Energy Holdup [p1]",
            "Energy Holdup [p2]",
            "Energy Accumulation [p1]",
            "Energy Accumulation [p2]",
            "Elemental Holdup [H]",
            "Elemental Holdup [He]",
            "Elemental Holdup [Li]",
            "Elemental Accumulation [H]",
            "Elemental Accumulation [He]",
            "Elemental Accumulation [Li]",
            "Elemental Transfer Term [H]",
            "Elemental Transfer Term [He]",
            "Elemental Transfer Term [Li]",
        ]

    assert len(dd["exprs"]) == 12
    for k in dd["exprs"].keys():
        assert k in [
            "Element Flow In [p1, H]",
            "Element Flow In [p1, He]",
            "Element Flow In [p1, Li]",
            "Element Flow In [p2, H]",
            "Element Flow In [p2, He]",
            "Element Flow In [p2, Li]",
            "Element Flow Out [p1, H]",
            "Element Flow Out [p1, He]",
            "Element Flow Out [p1, Li]",
            "Element Flow Out [p2, H]",
            "Element Flow Out [p2, He]",
            "Element Flow Out [p2, Li]",
        ]

    assert len(dd["params"]) == 0


@pytest.mark.unit
def test_reports():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)

    m.fs.cv = ControlVolume0DBlock(property_package=m.fs.pp, reaction_package=m.fs.rp)

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=True)
    m.fs.cv.add_reaction_blocks(has_equilibrium=True)
    m.fs.cv.add_material_balances(
        has_rate_reactions=True,
        has_equilibrium_reactions=True,
        has_phase_equilibrium=True,
    )
    m.fs.cv.add_energy_balances(has_heat_of_reaction=True, has_heat_transfer=True)
    m.fs.cv.add_momentum_balances(has_pressure_change=True)

    m.fs.cv.report()


@pytest.mark.unit
def test_dynamic_mass_basis():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=True, time_units=units.s)
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.rp = ReactionParameterTestBlock(property_package=m.fs.pp)
    m.fs.pp.basis_switch = 2
    m.fs.cv = ControlVolume0DBlock(
        property_package=m.fs.pp,
        reaction_package=m.fs.rp,
        dynamic=True,
    )

    m.fs.cv.add_geometry()
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_reaction_blocks(has_equilibrium=False)

    mb = m.fs.cv.add_phase_component_balances()

    assert isinstance(mb, Constraint)
    assert len(mb) == 8
    assert isinstance(m.fs.cv.phase_fraction, Var)
    assert isinstance(m.fs.cv.material_holdup, Var)
    assert isinstance(m.fs.cv.material_accumulation, Var)

    assert_units_consistent(m)
    assert_units_equivalent(m.fs.cv.material_holdup, units.kg)
    assert_units_equivalent(m.fs.cv.material_accumulation, units.kg / units.s)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
API for accessing core IDAES datasets

Usage, e.g., for the Pitzer(1984) data::

    from idaes.core.datasets import Pitzer
    pitzer = Pitzer()
    gibbs_data = pitzer.get_table("Standard G").data

"""
__authors__ = ["Dan Gunter (LBNL)"]
__author__ = __authors__[0]

from pyomo.common.deprecation import deprecation_warning


deprecation_warning(
    msg=(
        "idaes.core.datasets is no longer available due to the DMF being unsupported."
        " Visit https://github.com/IDAES/dmf for more information."
    ),
    version="2.6.0dev0",
    remove_in="2.6.0rc0",
)
